#include <linux/module.h>
#include <linux/param.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/malloc.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/segment.h>

/* In 2.2.3 /usr/include/linux/version.h includes 
 * a macro for this, but 2.0.35 doesn't - so I add 
 * it here if necessary. */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

/* Conditional compilation. LINUX_VERSION_CODE is 
 * the code (as per KERNEL_VERSION) of this version. */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>  /* for put_user */
#include <linux/poll.h>
#endif

// Delay in jiffies between servo movements.
#define DELAY HZ/10

static int security_major = 0;
static int timers = 0;
struct timer_list security_timer;
int position = 0,dir,stopped = 0, movement_detected = 0;

int countDown, fd,i;
long count, recv, avg;
unsigned char *buf,*tmp_uchptr,*original;
int portdata;
long k;
unsigned char countArray[]={0x80,0xf2,0x48,0x60,0x32,0x24,0x04,0xf0,0x00,0x20};
	

#define PWM_BASE    0x220
#define PWM_TIMER_0 0x220
#define PWM_TIMER_1 0x221
#define PWM_TIMER_2 0x222
#define PWM_CONTROL 0x223
#define SERVO_MIN   940
#define SECURITY_BASE 0x240
#define MAXDEGREES 100
#define NUM_POSITIONS SERVO_MIN/10 - 10
#define INFRARED 20
int value=SERVO_MIN;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
#define   copy_to_user  	memcpy_tofs
#define   copy_from_user        memcpy_fromfs
#endif

int turn(int degrees) {
  return SERVO_MIN + degrees * (int)((double)SERVO_MIN / MAXDEGREES);
}


static int security_open(struct inode *inode, struct file *filp){
  int minor = (MINOR(inode->i_rdev));

  if (minor > 1 && minor < 0) {
    printk("Error Security: The minor number should be either 0 or 1\n");
    return -EINVAL;
  }
  MOD_INC_USE_COUNT;
  return 0;
}

/* This function is called when a process closes the 
 * device file. It doesn't have a return value in 
 * version 2.0.x because it can't fail (you must ALWAYS
 * be able to close a device). In version 2.2.x it is 
 * allowed to fail - but we won't let it. 
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int security_release(struct inode *inode, struct file *filp){ 
#else 
     void security_release(struct inode *inode, struct file *filp){
#endif

  MOD_DEC_USE_COUNT;
  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t security_read(
    struct file *filp,
    char *buf,        /* The buffer to fill with data */
    size_t count,     /* The length of the buffer */
    loff_t *offset){  /* Our offset in the file */
#else
static int security_read(
    struct inode *inode, 
    struct file *filp,
    char *buf,        /* The buffer to fill with the data */ 
    int count         /* The length of the buffer,mustn't write beyond that! */
){
#endif

  int retval = 1;
  unsigned port = 0;
  unsigned char portdata;

  portdata = (unsigned char)stopped;
  stopped = 0;

  copy_to_user(buf, &portdata,retval);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t security_write(
    struct file *filp,
    const char *buf,
    size_t count,
    loff_t *offset
){
#else
int security_write(
    struct inode *inode,
    struct file *filp,
    const char *buf,
    int count
){
#endif

  int retval = count;
  unsigned port = 0;
  unsigned char outval;
  unsigned char *kbuf, *ptr;
  unsigned long j;
  movement_detected = 1;
  
  kbuf = (unsigned char *) kmalloc(count, GFP_KERNEL);
  if (!kbuf)
    return -ENOMEM;

  copy_from_user(kbuf, buf, count);
  ptr=kbuf;
  //  outval = *(ptr++);

  kfree(kbuf);
  return retval;

}


struct file_operations security_fops = {
  NULL,          /* i_bus_lseek   */
  security_read,
  security_write,
  NULL,          /* i_bus_readdir */
  NULL,       	 /* i_bus_select  */
  NULL,          /* i_bus_ioctl   */
  NULL,          /* i_bus_mmap    */
  security_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  NULL,   /* flush */
#endif
  security_release
};

 void automatic_mode(unsigned long param) {

   if (movement_detected) {
     movement_detected = 0;
     // Count back
     


     countDown = 9;
	
     while ( countDown >= 0) {
       portdata=countArray[countDown];

       outb(portdata, SECURITY_BASE);
       countDown--;

       // Check for key
       avg = 0;
       for (k=0;k<100;k++) {
	 for(i=0;inb(SECURITY_BASE) & 1 == 1;i++);
	 avg += i;
       }
       avg /= 100;

       if (avg < INFRARED) {
	 outb(0xFF,SECURITY_BASE);
	 for (k=0;k<2000;k++);
	 outb(0xFE,SECURITY_BASE);
	 break;
       }
       for (k=0;k<10000000;k++);
     }

     if (countDown < 0) {
       outb(0xff, SECURITY_BASE);
     }

   }

   outb(0xb2, PWM_CONTROL);
   outb(value & 0xFF, PWM_TIMER_2);
   outb(value >> 8 & 0xFF, PWM_TIMER_2);

   
   if (position == (NUM_POSITIONS)/2) {
     security_timer.expires = jiffies + 5*HZ;
     stopped = 1;
   } else {
     security_timer.expires = jiffies + DELAY;
   }

   if (position == NUM_POSITIONS) {
     dir = -1;
   } else if (position == 0) {
     dir = 1;
   }
   value += 10*dir;
   position += dir;

   security_timer.function = automatic_mode;
   add_timer(&security_timer);
 }


int init_module(void){

  int result, dyn_mjnum;

  result = check_region(PWM_BASE, 32);

  if (result) {
    printk(KERN_INFO "pwm: can't get I/O address 0x%x\n", PWM_BASE);
    return result;
  }
  request_region(PWM_BASE, 32, "pwm");

  result = check_region(SECURITY_BASE, 32);
  if (result) {
    printk(KERN_INFO "security: can't get I/O address 0x%x\n", SECURITY_BASE);
    return result;
  }
  request_region(SECURITY_BASE, 32, "security");
  outb(0xFE,SECURITY_BASE);

  printk("<1>pwm installed...\n");

  /* set up timer 0 to give periodic signals at 100 Hz
   * this signal is derived from the ISA CLK signal
   * which varies from machine to machine
   * on our lab machines, the frequency is eight and one third MHz
   */

  outb(0x34, PWM_CONTROL);
  outb(0, PWM_TIMER_0);
  outb(37, PWM_TIMER_0);
  printk("<1>timer 0 initialized...\n");

  /* set up timer 2 to half-scale model aircraft servo position
   * this is a one and one half millisecond pulse
   */
  
  outb(0xb2, PWM_CONTROL);
  outb((int)SERVO_MIN & 0xFF, PWM_TIMER_2);
  outb((int)SERVO_MIN >> 8 & 0xFF, PWM_TIMER_2);
  printk("<1>timer 2 initialized...\n");
    
  /* set up timer 1 also to center of servo range (as with above servo)
   *  this would be half-way
   */

  outb(0x72, PWM_CONTROL);
  outb((int)SERVO_MIN & 0xFF, PWM_TIMER_1);
  outb((int)SERVO_MIN >> 8 & 0xFF, PWM_TIMER_1);
  printk("<1>timer 1 initialized...\n");

  dyn_mjnum = register_chrdev(security_major, "security", &security_fops);
  if (security_major == 0) security_major = dyn_mjnum; /* dynamic */

  init_timer(&security_timer);
  security_timer.function = automatic_mode;
  security_timer.expires = jiffies + DELAY;
  add_timer(&security_timer);
  return 0;
}

void cleanup_module(void){
  int r;
  release_region(PWM_BASE, 32);
  release_region(SECURITY_BASE, 32);
  del_timer(&security_timer);

  r =  unregister_chrdev(security_major, "security");
    /* ERROR CHECKING */
  if (r==0) printk("<0>Security unloaded successfully!\n");
  if (r==-EINVAL) printk("<0>Invalid major number while unloading!\n"); 
  if (r==-EBUSY) printk("<0>Major number busy!");
}




